import sys
from pathlib import Path
from typing import BinaryIO, Union


def is_silk_data(raw: bytes) -> bool:
    if len(raw) > 10:
        offset = 0
        if raw[0] == 2:
            offset = 1
        if raw[offset:10] == b"#!SILK_V3":
            return True
    return False


def get_file(file: Union[str, BinaryIO]) -> BinaryIO:
    if isinstance(file, str):
        path = Path(file)
        if not path.is_file():
            raise FileNotFoundError(file)
        return open(file, "rb")
    elif isinstance(file, BinaryIO):
        return file
    else:
        raise TypeError(file)


def force_quit():
    import os
    import multiprocessing
    os.kill(multiprocessing.current_process().pid, 15)  # sigterm


def _play_sound(source: Union[str, bytes]):
    import winsound
    from threading import Thread

    t = Thread(
        target=winsound.PlaySound,
        name="PlayerThread",
        args=(source, winsound.SND_FILENAME if isinstance(source, str) else winsound.SND_MEMORY),
    )
    t.start()
    try:
        while True:
            t.join(0.5)
    except KeyboardInterrupt:
        print("Interrupt received")
        force_quit()


def play_audio(source: Union[str, bytes]):
    if sys.platform != "win32":
        raise RuntimeError("PlaySound only support windows")

    _play_sound(source)



import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
from os import cpu_count
from typing import Union, BinaryIO

from .coder import silkEncode, silkDecode
from .utils import get_file, is_silk_data
from .wav import Wave

_LOOP = asyncio.get_event_loop()
_EXECUTOR = ThreadPoolExecutor(cpu_count())


def encode(data: bytes, data_rate=24000, *, sample_rate=24000) -> bytes:
    if data[8:12] == b"WAVE":
        return silkEncode(
            Wave.wav2pcm(BytesIO(data)), sample_rate, data_rate
        )
    return silkEncode(data, sample_rate, data_rate)


def encode_file(target: Union[str, BinaryIO], data_rate=24000, *, sample_rate=24000) -> bytes:
    fd = get_file(target)
    try:
        return encode(fd.read(), data_rate, sample_rate=sample_rate)
    finally:
        fd.close()


def decode(silk_data: bytes, to_wav=False, *, sample_rate=24000) -> bytes:
    if is_silk_data(silk_data):
        if not to_wav:
            return silkDecode(silk_data, sample_rate)
        else:
            return Wave.pcm2wav(
                BytesIO(silkDecode(silk_data, sample_rate))
            )
    else:
        raise ValueError("Not a valid silk_data")


def decode_file(silk_file: Union[str, BinaryIO], to_wav=False, *, sample_rate=24000) -> bytes:
    fd = get_file(silk_file)
    try:
        return decode(fd.read(), to_wav, sample_rate=sample_rate)
    finally:
        fd.close()


async def async_encode(data: bytes, data_rate=24000, *, sample_rate=24000) -> bytes:
    return await _LOOP.run_in_executor(
        _EXECUTOR,
        functools.partial(
            encode,
            data,
            data_rate,
            sample_rate=sample_rate
        )
    )


async def async_decode(silk_data: bytes, to_wav=False, *, sample_rate=24000) -> bytes:
    return await _LOOP.run_in_executor(
        _EXECUTOR,
        functools.partial(
            decode,
            silk_data,
            sample_rate=sample_rate,
            to_wav=to_wav
        )
    )


async def async_encode_file(target: Union[str, BinaryIO], data_rate=24000, *, sample_rate=24000) -> bytes:
    return await _LOOP.run_in_executor(
        _EXECUTOR,
        functools.partial(
            encode_file,
            target,
            data_rate,
            sample_rate=sample_rate
        )
    )


async def async_decode_file(silk_file: Union[str, BinaryIO], to_wav=False, *, sample_rate=24000) -> bytes:
    return await _LOOP.run_in_executor(
        _EXECUTOR,
        functools.partial(
            decode_file,
            silk_file,
            sample_rate=sample_rate,
            to_wav=to_wav
        )
    )



import wave
from io import BytesIO
from typing import BinaryIO


class Wave:
    @staticmethod
    def pcm2wav(obj: BinaryIO, frame_rate=24000, *, ac=1) -> bytes:
        res = BytesIO()
        with wave.Wave_write(res) as wav:
            wav.setframerate(frame_rate)
            wav.setsampwidth(ac * 2)
            wav.setnchannels(ac)
            wav.writeframes(obj.read())
        return res.getvalue()

    @staticmethod
    def wav2pcm(obj: BinaryIO) -> bytes:
        res = BytesIO()
        with wave.Wave_read(obj) as wav:
            if wav.getcomptype() != "NONE":
                raise wave.Error("Unsupport with-compressed wave file")
            elif wav.getnchannels() == 1:
                while True:
                    bl = wav.readframes(512)
                    if bl:
                        res.write(bl)
                    else:
                        break
            else:
                channel_data = bytearray()
                while True:
                    frame = wav.readframes(1)
                    if frame:
                        channel_data.append(frame[0])
                    else:
                        break
                res.write(channel_data)
        return res.getvalue()



from .pysilk import encode, encode_file, decode, decode_file
from .pysilk import async_encode, async_encode_file, async_decode, async_decode_file

from .coder import __version__  # noqa

__all__ = [
    "encode",
    "encode_file",
    "decode",
    "decode_file",
    "async_encode",
    "async_encode_file",
    "async_decode",
    "async_decode_file"
]



import argparse
import time

from . import encode_file, decode_file
from .utils import get_file, play_audio
from .wav import Wave

parser = argparse.ArgumentParser("pysilk", description="encode/decode your silk file")
parser.add_argument("-r", "--rate", default=24000, type=int, help="set silk output datarate")
parser.add_argument("-p", "--play", action="store_true", help="play input file")
parser.add_argument("-q", "--quiet", action="store_true", help="reduce console output")
parser.add_argument("--sample-rate", default=24000, type=int, help="set pcm output samplerate")
parser.add_argument("input", action="store", type=str, help="input file path")
parser.add_argument("output", help="output file path", nargs="?", default="")


def get_suffix(path: str) -> str:
    if path:
        sp = path.rsplit(".", 1)
        if len(sp) == 1:
            raise ValueError("cannot parse suffix")
        elif sp[1] not in ("wav", "pcm", "silk"):
            raise TypeError("%s format not supported" % sp[1])
        else:
            return sp[1]


def log(*content_args):
    if not args.quiet:
        print(*content_args)


if __name__ == '__main__':
    st = time.time()
    args = parser.parse_args()
    i_suffix, o_suffix = get_suffix(args.input), get_suffix(args.output)
    if args.play:
        log("playing:", args.input)
        if i_suffix == "wav":
            play_audio(args.input)
        elif i_suffix == "pcm":
            play_audio(Wave.pcm2wav(get_file(args.input), args.sample_rate))
        elif i_suffix == "silk":
            play_audio(decode_file(args.input, to_wav=True))
    elif i_suffix == o_suffix or not o_suffix:
        log("nothing can do.")
    elif i_suffix == "pcm" and not args.sample_rate:
        raise ValueError("--sample-rate must be set")
    else:
        with open(args.output, "wb") as f:
            source = args.input
            if i_suffix == "wav" and o_suffix == "pcm":
                f.write(Wave.wav2pcm(get_file(source)))
            elif i_suffix == "pcm" and o_suffix == "wav":
                f.write(Wave.pcm2wav(get_file(source), args.sample_rate))
            elif i_suffix in ("pcm", "wav") and o_suffix == "silk":
                f.write(encode_file(source, args.rate))
            elif i_suffix == "silk" and o_suffix in ("pcm", "wav"):
                f.write(decode_file(source, to_wav=args.output == "wav"))
            else:
                print("Unknown operation:", i_suffix, "to", o_suffix)
    log(f"done, {round((time.time() - st) * 1000, 2)}ms used")



import glob
from distutils import log
from distutils.dep_util import newer_group
from distutils.errors import DistutilsSetupError
from sys import version_info

from setuptools import setup
from setuptools.command.test import test as tester

# Available at setup time due to pyproject.toml

try:
    from pybind11.setup_helpers import Pybind11Extension as Extension
    from pybind11.setup_helpers import build_ext
except ImportError:
    from setuptools import Extension
    from distutils.command.build_ext import build_ext

__version__ = "1.6.0"
basic_dependency = ["pybind11", "setuptools"]

if version_info.major != 3 or version_info.minor < 6:
    raise RuntimeError("pysilk only support python 3.6 or newer")


class CustomBuilder(build_ext):
    def build_extension(self, ext):
        sources = ext.sources
        if sources is None or not isinstance(sources, (list, tuple)):
            raise DistutilsSetupError(
                "in 'ext_modules' option (extension '%s'), "
                "'sources' must be present and must be "
                "a list of source filenames" % ext.name)
        sources = list(sources)
        ext_path = self.get_ext_fullpath(ext.name)
        depends = sources + ext.depends
        if not (self.force or newer_group(depends, ext_path, 'newer')):
            log.debug("skipping '%s' extension (up-to-date)", ext.name)
            return
        else:
            log.info("building '%s' extension", ext.name)

            # First, scan the sources for SWIG definition files (.i), run
        # SWIG on 'em to create .c files, and modify the sources list
        # accordingly.
        sources = self.swig_sources(sources, ext)

        # Next, compile the source code to object files. 

        # XXX not honouring 'define_macros' or 'undef_macros' -- the
        # CCompiler API needs to change to accommodate this, and I
        # want to do one thing at a time! 

        # Two possible sources for extra compiler arguments:
        #   - 'extra_compile_args' in Extension object
        #   - CFLAGS environment variable (not particularly
        #     elegant, but people seem to expect it and I
        #     guess it's useful)
        # The environment variable should take precedence, and
        # any sensible compiler will give precedence to later
        # command line args.  Hence we combine them in order:
        extra_args = ext.extra_compile_args or []

        macros = ext.define_macros[:]
        for undef in ext.undef_macros:
            macros.append((undef,))

            # split sources
        cxx_files = []
        for s in sources:  # type: str
            if s.endswith(".cpp"):
                sources.remove(s)
                cxx_files.append(s)

        c_build_args = []
        for e in extra_args:
            if not e.startswith("-std="):
                c_build_args.append(e)

        objects = self.compiler.compile(sources,
                                        output_dir=self.build_temp,
                                        macros=macros,
                                        include_dirs=ext.include_dirs,
                                        debug=self.debug,
                                        extra_postargs=c_build_args,
                                        depends=ext.depends)

        objects += self.compiler.compile(cxx_files,
                                         output_dir=self.build_temp,
                                         macros=macros,
                                         include_dirs=ext.include_dirs,
                                         debug=self.debug,
                                         extra_postargs=extra_args,
                                         depends=ext.depends)

        # XXX outdated variable, kept here in case third-part code
        # needs it.
        self._built_objects = objects[:]

        # Now link the object files together into a "shared object" --
        # of course, first we have to figure out all the other things
        # that go into the mix.
        if ext.extra_objects:
            objects.extend(ext.extra_objects)
        extra_args = ext.extra_link_args or []

        # Detect target language, if not provided
        language = ext.language or self.compiler.detect_language(sources)

        self.compiler.link_shared_object(
            objects, ext_path,
            libraries=self.get_libraries(ext),
            library_dirs=ext.library_dirs,
            runtime_library_dirs=ext.runtime_library_dirs,
            extra_postargs=extra_args,
            export_symbols=self.get_export_symbols(ext),
            debug=self.debug,
            build_temp=self.build_temp,
            target_lang=language)


class PyTest(tester):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        tester.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        tester.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        exit(
            pytest.main(self.pytest_args)
        )


def get_compile_file_list():
    files = glob.glob("src/silk/src/*.c")
    files.extend(["src/silk/_pysilk.cpp", "src/silk/codec.cpp"])
    return files


setup(
    version=__version__,
    requires=basic_dependency,
    tests_require=basic_dependency + ["pytest"],
    cmdclass={"test": PyTest, "build_ext": CustomBuilder},
    zip_safe=True,
    ext_modules=[
        Extension(
            "pysilk.coder", get_compile_file_list(),
            include_dirs=["src/silk/interface"],
            define_macros=[('VERSION_INFO', __version__)],
            extra_compile_args=["-std=c++11"]
        )
    ]
)
